(*
   This file is part of the Anoq SML Basis project. This project
   implements Anoq's proposals for some extensions to the
   Standard ML Basis Library, which will make the
   Basis Library interfaces more consistent.

   Copyright (C) 2003-2007 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.HardcoreProcessing.com/contact
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation;
   version 2.1 of the License.

   As a special exception, when making a "work that uses the library",
   you may compile this library directly into that work, without making
   it possible to relink that work with a modified version of this library.
   This is a relaxation of some of the requirements in section 6 of the
   GNU Lesser General Public License version 2.1.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License version 2.1 for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information:
   http://www.HardcoreProcessing.com/contact

   You should have received a copy of the GNU Lesser General Public
   License version 2.1 (WITHOUT the option of using a later version)
   along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*)

(* The main differences that this module has with
   the Array2 module are the following:
   1) tabulate[{l|r}] do not take the traversal argument
      (use the *Traversal functions instead)
   2) Some functions have an extra record-level for index or dimensions.
      This goes for: tabulate, sub and update, fold{l|r}i, mapi etc.
      This makes the signature consistent, but could give a performance
      hit on some compilers,
      e.g. if they fail to pass the nested record elements in registers.
   3) fromList now makes a one-dimensional array (use fromListList)
   4) app and appi etc. do not operate on regions any more, in accordance
      with the rest of the SML Basis Library filosophy of using
      a corresponding Array2DSlice module for that intead. 
*)
(* Warning: This implementation is completely untested! *)
(* Missing: map{l|r}[i], modify[i] *)
structure Array2D = (* Define a new and better Array2 module :) *)
  struct
    datatype traversal =
      XY (* Corresponds to RowMajor in Array2 *)
    | YX (* Corresponds to ColMajor in Array2 *)

    (* Example: On a 3x3 array, the elements are traversed as shown below,
                assuming that x extends to the right and y extends downwards:
       XY Order: YX Order:  Indices of the "table entries" on the left here:
       123       147        {x = 0, y = 0}, {x = 1, y = 0}, {x = 2, y = 0}
       456       258        {x = 0, y = 1}, {x = 1, y = 1}, {x = 2, y = 1}
       789       369        {x = 0, y = 2}, {x = 1, y = 2}, {x = 2, y = 2}
    *)
    
    type 'a t = ('a Array2.array) * Array2.traversal
    type index      = {x : int , y : int}
    type dimensions = {xs : int, ys : int}

    exception NotImplemented

    val defaultTraversal = XY

    (* The traversal order is stored along with the array.
       It is stored persistently though, so setTraversal makes a
       new array data value with the new traversal, but where
       the reference to the same imperative array is maintained.
       This allows for e.g. traversal of the same array by several
       threads at the same time with differing traversal orders. *)
    fun internalConvertTraversal XY =
        Array2.RowMajor
      | internalConvertTraversal YX =
        Array2.ColMajor

    fun setTraversal trav (arr, _) =
        (arr, internalConvertTraversal trav)
      
    fun getTraversal (_, Array2.RowMajor) =
        XY
      | getTraversal (_, Array2.ColMajor) =
        YX

    (* Conversion to and from the Array2 module, to increase usability *)
    fun toArray2 (arr, _) =
        arr

    fun fromArray2 arr =
        (arr, internalConvertTraversal defaultTraversal)

    fun create ({xs, ys} : dimensions, v) =
        (Array2.array (xs, ys, v),
         internalConvertTraversal defaultTraversal)

    fun tabulateTraversal traversal ({xs, ys}, f) =
        let
          fun fArray2(r, c) =
              f {x = c, y = r}

          val trav = internalConvertTraversal traversal
        in
          (Array2.tabulate trav (xs, ys, fArray2), trav)
        end

    fun tabulate args =
        tabulateTraversal defaultTraversal args

    fun empty () =
        tabulate ({xs = 0, ys = 0}, fn _ => raise Subscript)
          
    fun sub ((arr, _), {x, y}) =
        Array2.sub (arr, y, x)

    (* "This is one of them dreaded imperative functions..." *)
    fun update ((arr, _), {x, y}, v) =
        Array2.update (arr, y, x, v)

    fun dimensions (arr, _) =
        let
          val (rows, cols) = Array2.dimensions arr
        in
          {xs = cols, ys = rows}
        end

    fun count arr =
        let
          val {xs, ys} = dimensions arr
        in
          xs * ys
        end
    
    fun getXs ((arr, _), y) =
        Array2.row(arr, y)

    fun getYs ((arr, _), x) =
        Array2.column(arr, x)
    
    fun foldli f init (arr, trav) =
          Array2.foldi trav
                       (fn (r, c, v, acc) => f ({x = c, y = r}, v, acc))
                       init {base = arr, col = 0, row = 0,
                             ncols = NONE, nrows = NONE}

    fun foldl f init arr =
          foldli (fn (_, v, acc) => f (v, acc)) init arr

    fun foldri f init (arr, trav) =
        let
          val {xs, ys} = dimensions (arr, trav)

          fun iterXy (i as {x, y}, acc) =
              if x >= 0 then
                if y >= 0 then
                  iterXy
                    ({x = x - 1, y = y}, f(i, sub((arr, trav), i), acc))
                else
                  acc
              else
                iterXy({x = xs - 1, y = y - 1}, acc)

          fun iterYx (i as {x, y}, acc) =
              if y >= 0 then
                if x >= 0 then
                  iterYx
                    ({x = x, y = y - 1}, f(i, sub((arr, trav), i), acc))
                else
                  acc
              else
                iterYx({x = x - 1, y = ys - 1}, acc)

        in
          case trav of
            Array2.RowMajor =>
              iterXy({x = xs - 1, y = ys - 1}, init)
          | Array2.ColMajor =>
              iterYx({x = xs - 1, y = ys - 1}, init)
        end

    fun foldr f init arr =
        foldri (fn (_, v, acc) => f (v, acc)) init arr

    fun foldsepli sep f init l =
          let
            fun foldF (i as {x, y}, e, acc) =
                  if x <> 0 orelse y <> 0 then
                    f (i, e, sep acc)
                  else
                    f (i, e, acc)
          in
            foldli foldF init l
          end

    fun foldsepri sep f init l =
          let
            fun foldF (i as {x, y}, e, acc) =
                  if x <> 0 orelse y <> 0 then
                    sep (f (i, e, acc))
                  else
                    f (i, e, acc)
          in
            foldri foldF init l
          end

    fun foldsepl sep f acc str =
          foldsepli sep (fn (_, e, acc) => f (e, acc)) acc str

    fun foldsepr sep f acc str =
          foldsepri sep (fn (_, e, acc) => f (e, acc)) acc str

    (* Hey, look how nice! Except for the traversal-thing,
       this implementation is exactly the same as for Array.map :) *)
    fun map f (arr : 'a t) =
          tabulateTraversal (getTraversal arr)
            (dimensions arr,
             fn i => f (sub(arr, i)))

    (* Hey, look how nice! Except for the traversal-thing,
       this implementation is exactly the same as for Array.mapi :) *)
    fun mapi f (arr : 'a t) =
          tabulateTraversal (getTraversal arr)
            (dimensions arr,
             fn i => f (i, sub(arr, i)))

    fun appi f (arr, trav) =
          Array2.appi
            trav
            (fn (r, c, v) => f ({x = c, y = r}, v))
            {base = arr, col = 0, row = 0,
             ncols = NONE, nrows = NONE}

    fun app f arr =
          appi (fn (_, v) => f v) arr

    fun tabulatelTraversal traversal ({xs, ys}, f, init) =
        let
          val rAcc = ref init

          fun fArray2(r, c) =
              let
                val (e, acc) = f({x = c, y = r}, !rAcc)
                val _ = rAcc := acc
              in
                e
              end

          val trav = internalConvertTraversal traversal
        in
          ((Array2.tabulate trav (xs, ys, fArray2), trav),
           !rAcc)
        end

    fun tabulatel args =
        tabulatelTraversal defaultTraversal args

    (* This one is a bit tricky and probably not the most efficient... *)
    fun tabulaterTraversal traversal (dim as {xs, ys}, f, init) =
        if xs > 0 andalso ys > 0 then
          let
            (* Apply function first time, to get a "dummy element" for array *)
            val (e0, init') = f({x = xs - 1, y = ys - 1}, init)

            fun fFirstPass(i, acc) =
                (e0, i::acc)
       
            (* Tabulate the array to hold the resulting elements,
               while accumulating list of indices to fold over. *)
            val (arr, l) = tabulatelTraversal traversal (dim, fFirstPass, nil)

            fun fUserPass(i, acc) =
                let
                  val (e, acc') = f(i, acc)
                  val _ = update(arr, i, e)
                in
                  acc'
                end

            (* Do the "fold" operation of the user-supplied function,
               skipping last index, since we have already applied f on that. *)
            val result = List.foldl fUserPass init' (tl l)
          in
            (arr, result)
          end
        else
          (* No elements, so traversal order does not matter *)
          tabulatelTraversal traversal (dim, f, init)

    fun tabulater args =
        tabulaterTraversal defaultTraversal args

    fun append (leftArr, rightArr) =
          (* FIXME: I guess this should concatenate two
             arrays either "beside" each other or
             "on top of" each other. Maybe according
             to what makes most sense with the given traversal? *)
          raise NotImplemented

    fun concat arrs =
        (* FIXME: Several appends? *)
        raise NotImplemented

    local
      fun findiF f (i, e, NONE) =
            if f (i, e) then
              SOME e
            else
              NONE
        | findiF f (_, _, acc as (SOME _)) =
            acc
    in
      fun findi f arr =
            foldli (findiF f) NONE arr
    end

    fun find f arr =
        findi (fn (_, e) => f e) arr

    fun existsi f arr =
          foldli (fn (i, e, acc) => acc orelse (f (i, e))) false arr

    fun exists f arr =
          existsi (fn (_, e) => f e) arr

    fun fromListList ll =
          (Array2.fromList ll, internalConvertTraversal defaultTraversal)

    fun fromList l =
          (* FIXME: This may be a problem? *)
          fromListList [l]

    fun toList arr =
          foldr op:: nil arr

    fun rev arr =
          (* FIXME: What should rev do? Converting over a list changes layout *)
          (* fromList (foldl op:: nil arr) *)
          raise NotImplemented

    fun lookup arr (index as {x, y}) =
          (* It really ought to be sub which is implemented in terms
             of lookup, but we already have sub, so I haven't bothered.
             Probably requires compiler support. *)
          let
            val {xs, ys} = dimensions arr
          in
            if x >= 0 andalso x < xs andalso
               y >= 0 andalso y < ys then
              SOME (sub (arr, index))
            else
              NONE
          end

    (* Functional update. This overwrites the element at index.
       This is different from Vector, Array and List. *)
    fun insert index newV arr =
          mapi (fn (i, v) => if i = index then newV else v) arr

    fun delete index arr =
          (* I don't think this operation makes sense for Array2D *)
          raise NotImplemented

    fun substitute index f arr =
          mapi (fn (i, v) => if i = index then f v else v) arr

    (* More functional operations: These reduce the collection dimensions
       and do not make much sense for Array2. *)
    fun filter p t =
          raise NotImplemented

    fun filteri p t =
          raise NotImplemented


    fun splitAt (t, i) =
          raise NotImplemented


    fun splitl p t =
          raise NotImplemented

    fun splitli p t =
          raise NotImplemented

    fun splitr p t =
          raise NotImplemented

    fun splitri p t =
          raise NotImplemented

    fun takel p t =
          raise NotImplemented

    fun takeli p t =
          raise NotImplemented

    fun taker p t =
          raise NotImplemented

    fun takeri p t =
          raise NotImplemented

    fun dropl p t =
          raise NotImplemented

    fun dropli p t =
          raise NotImplemented

    fun dropr p t =
          raise NotImplemented

    fun dropri p t =
          raise NotImplemented

    (* triml/r require that dimensions is "negatable" in some sense,
       which should be the case for almost any sensible dimensions type. *)
    fun triml d t =
          raise NotImplemented

    fun trimr d t =
          raise NotImplemented


    fun keepl d t =
          raise NotImplemented

    fun keepr d t =
          raise NotImplemented
  end
