(*
   This file is part of the Anoq SML Basis project. This project
   implements Anoq's proposals for some extensions to the
   Standard ML Basis Library, which will make the
   Basis Library interfaces more consistent.

   Copyright (C) 2003-2007 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.HardcoreProcessing.com/contact
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation;
   version 2.1 of the License.

   As a special exception, when making a "work that uses the library",
   you may compile this library directly into that work, without making
   it possible to relink that work with a modified version of this library.
   This is a relaxation of some of the requirements in section 6 of the
   GNU Lesser General Public License version 2.1.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License version 2.1 for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information:
   http://www.HardcoreProcessing.com/contact

   You should have received a copy of the GNU Lesser General Public
   License version 2.1 (WITHOUT the option of using a later version)
   along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*)

(* An integer module with forl and forr functions. *)
structure Int =
  struct
    open Int

    (* The forl and forr functions could be curried, but one
       may be likely to forget the last init parameter,
       if using it for a side-effecting loop! *)
    fun forl (starti, iters, stepi, f, init) =
          if iters >= 0 then
            let
              val endi = starti + stepi * iters
                
              (* The inner loop, which should be efficient.
                 Only one live variable for counting: i. *)
              fun loopUp (i, acc) =
                    if i < endi then
                      loopUp (i + stepi, f (i, acc))
                    else
                      acc

              (* The inner loop for when stepi < 0 *)
              fun loopDown (i, acc) =
                    if i > endi then
                      loopDown (i + stepi, f (i, acc))
                    else
                      acc
            in
              if stepi > 0 then
                loopUp (starti, init)
              else if stepi < 0 then
                loopDown (starti, init)
              else
                raise Subscript
            end
          else
            raise Subscript

    fun forr (starti, iters, stepi, f, init) =
          if iters >= 0 then 
            let
              (* In forr the loop just counts backwards,
                 starting at the last iteration. Notice: iters - 1 *)
              val endi = starti + stepi * (iters - 1)
              
              (* The inner loop, which should be efficient.
                 Only one live variable for counting: i. *)
              fun loopUp (i, acc) =
                    (* Notice: i >= starti *)
                    if i >= starti then
                      loopUp (i - stepi, f (i, acc))
                    else
                      acc

              (* The inner loop for when stepi < 0 *)
              fun loopDown (i, acc) =
                    (* Notice: i <= starti *)
                    if i <= starti then
                      loopDown (i - stepi, f (i, acc))
                    else
                      acc
            in
              if stepi > 0 then
                loopUp (endi, init)
              else if stepi < 0 then
                loopDown (endi, init)
              else
                raise Subscript
            end
          else
            raise Subscript
  end
