(*
   This file is part of the Anoq SML Basis project. This project
   implements Anoq's proposals for some extensions to the
   Standard ML Basis Library, which will make the
   Basis Library interfaces more consistent.

   Copyright (C) 2003-2007 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.HardcoreProcessing.com/contact
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation;
   version 2.1 of the License.

   As a special exception, when making a "work that uses the library",
   you may compile this library directly into that work, without making
   it possible to relink that work with a modified version of this library.
   This is a relaxation of some of the requirements in section 6 of the
   GNU Lesser General Public License version 2.1.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License version 2.1 for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information:
   http://www.HardcoreProcessing.com/contact

   You should have received a copy of the GNU Lesser General Public
   License version 2.1 (WITHOUT the option of using a later version)
   along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*)

(* A real module with forl and forr functions. *)
structure Real =
  struct
    open Real

    (* The forl and forr functions could be curried, but one
       may be likely to forget the last init parameter,
       if using it for a side-effecting loop! *)
    fun forl (starti : real, iters : int, stepi : real, f, init) =
          if Int.>=(iters, 0) then
            let
              (* The inner loop, which should be efficient.
                 We need two live variables just for counting though.
                 iters controls the number of iterations, without floating
                 point inaccurracy problems, and i is the index. *)
              fun loop (i : real, iters : int, acc) =
                    if Int.>(iters, 0) then
                      loop (i + stepi, Int.-(iters, 1), f (i, acc))
                    else
                      acc
            in
              if stepi > 0.0 orelse stepi < 0.0 then
                loop (starti, iters, init)
              else
                raise Subscript
            end
          else
            raise Subscript

    (* forr is not tail-recursive and thus uses stack space!
       This is necessary to guarantee that the same index values are
       traversed, in the presence of floating point inaccuracy problems. *)
    fun forr (starti : real, iters : int, stepi : real, f, init) =
          if Int.>=(iters, 0) then
            let
              (* The inner loop, which should be efficient.
                 We need two live variables just for counting though
                 and it is not tail-recursive.
                 iters controls the number of iterations, without floating
                 point inaccurracy problems, and i is the index. *)
              fun loop (i : real, iters : int) =
                    if Int.>(iters, 0) then
                      f(i, loop (i + stepi, Int.-(iters, 1)))
                    else
                      init
            in
              if stepi > 0.0 orelse stepi < 0.0 then
                loop (starti, iters)
              else
                raise Subscript
            end
          else
            raise Subscript

  end
