(*
   This file is part of the Anoq SML Basis project. This project
   implements Anoq's proposals for some extensions to the
   Standard ML Basis Library, which will make the
   Basis Library interfaces more consistent.

   Copyright (C) 2003-2007 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.HardcoreProcessing.com/contact
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation;
   version 2.1 of the License.

   As a special exception, when making a "work that uses the library",
   you may compile this library directly into that work, without making
   it possible to relink that work with a modified version of this library.
   This is a relaxation of some of the requirements in section 6 of the
   GNU Lesser General Public License version 2.1.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License version 2.1 for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information:
   http://www.HardcoreProcessing.com/contact

   You should have received a copy of the GNU Lesser General Public
   License version 2.1 (WITHOUT the option of using a later version)
   along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*)

structure Vector = (* Define a new and better Vector module :) *)
  struct
    open Vector (* The current implementation of Vector *)

    (* Make sure that these functions are the SML Basis 2002 compatible ones *)
    open AnoqVector2002

    type index = int
    type dimensions = int
    type 'a t = 'a vector

    (* length could be deprecated *)
    val dimensions = length
    val count      = dimensions

    fun empty () =
          tabulate (0, fn _ => raise Subscript)
          
    fun create (dim, e) =
          tabulate (dim, fn _ => e)

    fun tabulatel (dim, f, init) =
          let
            val (l, result) = List.tabulatel (dim, f, init)
          in
            (fromList l, result)
          end

    fun tabulater (dim, f, init) =
          let
            val (l, result) = List.tabulater (dim, f, init)
          in
            (fromList l, result)
          end

    fun rev vec =
          fromList (foldl op:: nil vec)

    fun toList vec =
          foldr op:: nil vec

    fun mapli f init vec =
          let
            fun foldF (i, e, (resultarr, acc)) =
                 let
                   val (e', acc') = f (i, e, acc)
                 in
                   (e'::resultarr, acc')
                 end

            val (l, acc') = foldli foldF (nil, init) vec
          in
            (Vector.fromList (List.rev l), acc')
          end

    fun mapri f init vec =
          let
            fun foldF (i, e, (resultarr, acc)) =
                 let
                   val (e', acc') = f (i, e, acc)
                 in
                   (e'::resultarr, acc')
                 end

            val (l, acc') = foldri foldF (nil, init) vec
          in
            (Vector.fromList l, acc')
          end

    fun mapl f init vec =
          mapli (fn (_, e, acc) => f (e, acc)) init vec

    fun mapr f init vec =
          mapri (fn (_, e, acc) => f (e, acc)) init vec

    fun foldsepli sep f init l =
          let
            fun foldF (i, e, acc) =
                  if i > 0 then
                    f (i, e, sep acc)
                  else
                    f (i, e, acc)
          in
            foldli foldF init l
          end

    fun foldsepri sep f init l =
          let
            fun foldF (i, e, acc) =
                  if i > 0 then
                    sep (f (i, e, acc))
                  else
                    f (i, e, acc)
          in
            foldri foldF init l
          end

    fun foldsepl sep f acc str =
          foldsepli sep (fn (_, e, acc) => f (e, acc)) acc str

    fun foldsepr sep f acc str =
          foldsepri sep (fn (_, e, acc) => f (e, acc)) acc str

    (*
    fun append (leftVec, rightVec) =
          let
            val lDim = dimensions leftVec
          in
            tabulate
              (lDim + dimensions rightVec,
               fn i => if i < lDim then
                         sub (leftVec, i)
                       else
                         sub (rightVec, i - lDim)
              )
          end
    *)
    
    fun append (l, r) =
          concat [l, r]
    
    local
      fun findF f (e, NONE) =
            if f e then
              SOME e
            else
              NONE
        | findF f (_, acc) =
            (* acc is: SOME e *)
            acc

      fun findiF f (i, e, NONE) =
            if f (i, e) then
              SOME e
            else
              NONE
        | findiF f (_, _, acc) =
            (* acc is: SOME e *)
            acc
    in
      fun find f vec =
            foldl (findF f) NONE vec

      fun findi f vec =
            foldli (findiF f) NONE vec
    end

    fun exists f vec =
          foldl (fn (e, acc) => acc orelse (f e)) false vec

    fun existsi f vec =
          foldli (fn (i, e, acc) => acc orelse (f (i, e))) false vec

    fun lookup vec index =
          (* It really ought to be sub which is implemented in terms
             of lookup, but we already have sub, so I haven't bothered. *)
          if index >= 0 andalso index < dimensions vec then
            SOME (sub (vec, index))
          else
            NONE

    fun insert index v vec =
          tabulate
            (dimensions vec + 1,
             fn i => if i < index then
                       sub (vec, i)
                     else if i = index then
                       v
                     else
                       sub (vec, i - 1)
            )

    fun delete index vec =
          if index >= 0 then
            tabulate
               (dimensions vec - 1,
                fn i => if i < index then
                          sub (vec, i)
                        else
                          sub (vec, i + 1)
               )
          else
             (* If index < 0 then the tabulated function is not called,
                so we have to explicitly make this range check *)
             raise Subscript

    fun substitute index f vec =
          mapi (fn (i, v) => if i = index then f v else v) vec

    fun filteri p l =
          fromList
            (List.rev
              (foldli
                 (fn (i, e, acc) => if p (i, e) then e :: acc else acc)
                 nil l
              )
            )

    fun filter p l =
          fromList
            (List.rev
              (foldl
                 (fn (e, acc) => if p e then e :: acc else acc)
                 nil l
              )
            )

    fun triml dim vec =
          let
            val d = dimensions vec
          in
            if dim >= 0 andalso dim <= d then
              tabulate (dimensions vec - dim, fn i => sub (vec, dim + i))
            else
              raise Subscript
          end

    fun trimr dim vec =
          let
            val d = dimensions vec
          in
            if dim >= 0 andalso dim <= d then
              tabulate (dimensions vec - dim, fn i => sub (vec, i))
            else
              raise Subscript
          end

    fun keepl dim vec =
          let
            val d = dimensions vec
          in
            if dim >= 0 andalso dim <= d then
              tabulate (dim, fn i => sub (vec, i))
            else
              raise Subscript
          end

    fun keepr dim vec =
          let
            val d = dimensions vec
            val offset = d - dim
          in
            if dim >= 0 andalso dim <= d then
              tabulate (dim, fn i => sub (vec, offset + i))
            else
              raise Subscript
          end

    fun splitAt (vec, i) =
          (keepl i vec,
           triml i vec)

    fun splitl p l =
          let
            fun splitlF (e, (left, right, continue)) =
                  if continue andalso p e then
                    (e::left, right, true)
                  else
                    (left, e::right, false)

            val (left, right, _) =
                  foldl splitlF (nil, nil, true) l
          in
            (fromList (List.rev left),
             fromList (List.rev right))
          end

    fun splitli p l =
          let
            fun splitliF (i, e, (left, right, continue)) =
                  if continue andalso p (i, e) then
                    (e::left, right, true)
                  else
                    (left, e::right, false)

            val (left, right, _) =
                  foldli splitliF (nil, nil, true) l
          in
            (fromList (List.rev left),
             fromList (List.rev right))
          end

    fun splitr p l =
          let
            fun splitrF (e, (left, right, continue)) =
                  if continue andalso p e then
                    (left, e::right, true)
                  else
                    (e::left, right, false)

            val (left, right, _) =
                  foldr splitrF (nil, nil, true) l
          in
            (fromList left,
             fromList right)
          end

    fun splitri p l =
          let
            fun splitriF (i, e, (left, right, continue)) =
                  if continue andalso p (i, e) then
                    (left, e::right, true)
                  else
                    (e::left, right, false)

            val (left, right, _) =
                  foldri splitriF (nil, nil, true) l
                    
          in
            (fromList left,
             fromList right)
          end

    (* The take and drop functions can always be implemented by split. *)
    fun takel p l =
          #1 (splitl p l)

    fun takeli p l =
          #1 (splitli p l)

    fun taker p l =
          #2 (splitr p l)

    fun takeri p l =
          #2 (splitri p l)
    
    fun dropl p l =
          #2 (splitl p l)

    fun dropli p l =
          #2 (splitli p l)

    fun dropr p l =
          #1 (splitr p l)

    fun dropri p l =
          #1 (splitri p l)
  end
